
/* ************************************************************************* */
/*
 *	gen_test_data.c
 *
 *	(C) 2010 VOCAL Technologies, Ltd.
 *
 *	This software is intended for use as a reference/standard implementation,
 *	and has been placed into the public domain by the copyright holder:
 *
 *	VOCAL Technologies, Ltd.
 *	90A John Muir Drive.
 *	Buffalo, NY  14228
 *
 *	Maintainer:  Dennis Towne
 */
/* ************************************************************************* */

#include <stdlib.h>
#include <string.h>

#include "system.h"
#include "uuid_hash.h"
#include "util_tests.h"

#define SMALL_ARCHIVE_RECORD_COUNT  20000

#define NUMBER_MALE_NAMES     200
#define NUMBER_FEMALE_NAMES   200
#define MAX_NAME_LEN          100


/* Note - these names were obtained from the US Govt. Social Security web site, from */
/* the 'most commonly used 200 male and female names in the 1990-1999 decade' page. */

static const char *male_names[NUMBER_MALE_NAMES] = {
	"Michael", "Christopher", "Matthew", "Joshua", "Jacob", "Nicholas", "Andrew", "Daniel", "Tyler", "Joseph",
	"Brandon", "David", "James", "Ryan", "John", "Zachary", "Justin", "William", "Anthony", "Robert",
	"Jonathan", "Austin", "Alexander", "Kyle", "Kevin", "Thomas", "Cody", "Jordan", "Eric", "Benjamin",
	"Aaron", "Christian", "Samuel", "Dylan", "Steven", "Brian", "Jose", "Timothy", "Nathan", "Adam",
	"Richard", "Patrick", "Charles", "Sean", "Jason", "Cameron", "Jeremy", "Mark", "Stephen", "Jesse",
	"Juan", "Alex", "Travis", "Jeffrey", "Ethan", "Caleb", "Luis", "Jared", "Logan", "Hunter",
	"Trevor", "Bryan", "Evan", "Paul", "Taylor", "Kenneth", "Connor", "Dustin", "Noah", "Carlos",
	"Devin", "Gabriel", "Ian", "Nathaniel", "Gregory", "Derek", "Corey", "Jesus", "Scott", "Bradley",
	"Dakota", "Antonio", "Marcus", "Blake", "Garrett", "Edward", "Luke", "Shawn", "Peter", "Seth",
	"Mitchell", "Adrian", "Victor", "Miguel", "Shane", "Chase", "Isaac", "Spencer", "Lucas", "Jack",
	"Tanner", "Angel", "Vincent", "Isaiah", "Dalton", "Brett", "George", "Alejandro", "Elijah", "Cory",
	"Cole", "Joel", "Erik", "Jake", "Mason", "Jorge", "Dillon", "Raymond", "Colton", "Ricardo",
	"Casey", "Francisco", "Brendan", "Devon", "Keith", "Colin", "Wesley", "Phillip", "Oscar", "Julian",
	"Johnathan", "Eduardo", "Chad", "Donald", "Bryce", "Ronald", "Alec", "Dominic", "Grant", "Martin",
	"Henry", "Mario", "Xavier", "Manuel", "Alan", "Derrick", "Frank", "Tristan", "Collin", "Omar",
	"Jeremiah", "Jackson", "Troy", "Edgar", "Javier", "Douglas", "Clayton", "Jonathon", "Nicolas", "Andre",
	"Maxwell", "Ivan", "Philip", "Levi", "Sergio", "Roberto", "Darius", "Andres", "Cristian", "Hector",
	"Fernando", "Drew", "Curtis", "Gary", "Riley", "Johnny", "Max", "Dennis", "Malik", "Wyatt",
	"Cesar", "Edwin", "Gavin", "Preston", "Marco", "Ruben", "Allen", "Calvin", "Mathew", "Randy",
	"Brent", "Jerry", "Hayden", "Alexis", "Parker", "Brady", "Tony", "Pedro", "Craig", "Larry"
};

static const char *female_names[NUMBER_FEMALE_NAMES] = {
	"Jessica", "Ashley", "Emily", "Sarah", "Samantha", "Amanda", "Brittany", "Elizabeth", "Taylor", "Megan",
	"Hannah", "Kayla", "Lauren", "Stephanie", "Rachel", "Jennifer", "Nicole", "Alexis", "Victoria", "Amber",
	"Alyssa", "Courtney", "Rebecca", "Danielle", "Jasmine", "Brianna", "Katherine", "Alexandra", "Madison", "Morgan",
	"Melissa", "Michelle", "Kelsey", "Chelsea", "Anna", "Kimberly", "Tiffany", "Olivia", "Mary", "Christina",
	"Allison", "Abigail", "Sara", "Shelby", "Heather", "Haley", "Maria", "Kaitlyn", "Laura", "Erin",
	"Andrea", "Natalie", "Jordan", "Brooke", "Julia", "Emma", "Vanessa", "Erica", "Sydney", "Kelly",
	"Kristen", "Katelyn", "Marissa", "Amy", "Crystal", "Paige", "Cassandra", "Gabrielle", "Katie", "Caitlin",
	"Lindsey", "Destiny", "Kathryn", "Jacqueline", "Shannon", "Jenna", "Angela", "Savannah", "Mariah", "Alexandria",
	"Sierra", "Alicia", "Briana", "Miranda", "Jamie", "Catherine", "Brittney", "Breanna", "Grace", "Monica",
	"Sabrina", "Madeline", "Caroline", "Molly", "Erika", "Mackenzie", "Leah", "Diana", "Whitney", "Cheyenne",
	"Bailey", "Christine", "Meghan", "Lindsay", "Cynthia", "Angelica", "Margaret", "Kaitlin", "Alexa", "Hailey",
	"Veronica", "Melanie", "Bianca", "Autumn", "Ariel", "Kristin", "Bethany", "Lisa", "Kristina", "Holly",
	"Leslie", "Casey", "Chloe", "April", "Julie", "Claire", "Kaylee", "Brenda", "Kathleen", "Rachael",
	"Karen", "Sophia", "Patricia", "Gabriela", "Kendra", "Dominique", "Kara", "Desiree", "Ana", "Tara",
	"Michaela", "Brandi", "Carly", "Kylie", "Karina", "Adriana", "Valerie", "Caitlyn", "Natasha", "Hayley",
	"Rebekah", "Jocelyn", "Cassidy", "Jade", "Gabriella", "Makayla", "Daisy", "Alison", "Jillian", "Audrey",
	"Faith", "Angel", "Nancy", "Dana", "Krystal", "Alejandra", "Ariana", "Summer", "Mikayla", "Isabella",
	"Raven", "Katrina", "Kiara", "Sandra", "Meagan", "Kirsten", "Lydia", "Chelsey", "Zoe", "Monique",
	"Claudia", "Mallory", "Joanna", "Deanna", "Isabel", "Ashlee", "Felicia", "Marisa", "Mercedes", "Mckenzie",
	"Jasmin", "Krista", "Yesenia", "Diamond", "Evelyn", "Selena", "Cindy", "Brandy", "Gina", "Mia"
};


/******************************************************************************/
/* pseudo random number generator used for deterministic data generation */
/******************************************************************************/

#define MYRAND_MAX  32768

static unsigned long random_internal_state = 331350945;

/* NOTE - returns a deterministic PRN ranging from 0 to 32767 */
int myrand(void) {
    random_internal_state = random_internal_state * 1103515245 + 12345;
    return((unsigned)(random_internal_state/65536) % MYRAND_MAX);
}

void mysrand(unsigned seed) {
    random_internal_state = seed;
}


/******************************************************************************/
/* routines for generating simulated patient data records */
/******************************************************************************/

static int small_archive_generate = 1;

static int year_base = 1910;
static int month_base = 1;
static int day_base = 1;

int create_unique_ssn(void) {
	static int curr_ssn_block = 111222;
	static int curr_ssn_counter = 0;	/* start with ssn 111222000 */

	/* allocate SSN's in blocks of 1000 correlated to birth date */
	curr_ssn_counter++;
	if (curr_ssn_counter >= 1000) {
		curr_ssn_counter = 0;
		curr_ssn_block += 8009;
		if (curr_ssn_block >= 1000000) {
			curr_ssn_block -= 900000;
		}

		/* in small archive mode, when we switch to a new SSN block, we also pick */
		/* a new random date period for those SSNs */
		if (small_archive_generate) {
			year_base  = myrand() * 100 / MYRAND_MAX + 1910;
			month_base = myrand() *  12 / MYRAND_MAX +  1;
			day_base   = myrand() *  31 / MYRAND_MAX +  1;
		}
	}

	return ((curr_ssn_block * 1000) + curr_ssn_counter);
}

int create_patient_record(struct test_data *td) {
	static int day_patient_counter = 0;
	int i;
	int curr_ssn = create_unique_ssn();

	/* allocate 500 patients per day for every day since 1910 */
	/* this would be a total of around 17 million records */
	day_patient_counter++;
	if (day_patient_counter > 500) {
		day_patient_counter = 0;
		day_base++;
	}
	/* note that this will create some bogus days of year, eg Feb 31 */
	if (day_base > 31) {
		day_base = 1;
		month_base++;
	}
	if (month_base > 12) {
		month_base = 1;
		year_base++;
	}
	if (year_base > 2010) {
		return 0;
	}

	if (myrand() & 1) {
		td->patient_gender = UUID_GENDER_FEMALE;
		i = myrand() * NUMBER_FEMALE_NAMES / MYRAND_MAX;
		strxcpy(td->patient_name, female_names[i], sizeof(td->patient_name));
	}
	else {
		td->patient_gender = UUID_GENDER_MALE;
		i = myrand() * NUMBER_MALE_NAMES / MYRAND_MAX;
		strxcpy(td->patient_name, male_names[i], sizeof(td->patient_name));
	}

	td->patient_ssn    = curr_ssn;
	td->patient_year   = year_base;
	td->patient_month  = month_base;
	td->patient_day    = day_base;

	/* add policy holder data half of the time - for completeness */
	if (myrand() & 1) {
		if (myrand() & 1) {
			td->policy_gender = UUID_GENDER_FEMALE;
			i = myrand() * NUMBER_FEMALE_NAMES / MYRAND_MAX;
			strxcpy(td->policy_name, female_names[i], sizeof(td->policy_name));
		}
		else {
			td->policy_gender = UUID_GENDER_MALE;
			i = myrand() * NUMBER_MALE_NAMES / MYRAND_MAX;
			strxcpy(td->policy_name, male_names[i], sizeof(td->policy_name));
		}

		td->policy_ssn   = create_unique_ssn();
		td->policy_year  = myrand() * 100 / MYRAND_MAX + 1910;
		td->policy_month = myrand() *  12 / MYRAND_MAX +  1;
		td->policy_day   = myrand() *  31 / MYRAND_MAX +  1;
	}

	/* destroy the SSN on the records at random */
	if (myrand() & 1) {
		td->patient_ssn  = 0;
	}
	if (myrand() & 1) {
		td->policy_ssn   = 0;
	}

	/* NOTE - the above scheme will produce records with redundant/extra information
	 * in them.  The only valid input types to the algorithm are:
	 *
	 * 1) patient ssn available - use only patient data
	 *
	 * 2) patient ssn not available, policy ssn available - use both patient and
	 *    policy holder info, use policy holder SSN
	 *
	 * 3) neither ssn available - use both patient and policy holder info as in #2,
	 *    use a zero policy holder ssn
	 */

	return 1;
}


/******************************************************************************/
/* uuid display/generate helpers */
/******************************************************************************/

void generate_and_print_uuid(FILE *stream, struct test_data *td) {
	char uuid[UUID_HASH_STRING_LENGTH + 1];
	char temp;

	generate_test_uuid(td, uuid);

	fprintf(stream, "patient_uuid          %s\n", uuid);

	temp = uuid[0];
	if ((temp >= '0' && temp <= '9') || (temp >= 'a' && temp <= 'f')) {
		if (td->patient_ssn) {
			printf("Error in generate_and_print_uuid(), uuid_create() generated bad leading digit for patient SSN present.\n");
		}
	}
	else {
		if (!td->patient_ssn) {
			printf("Error in generate_and_print_uuid(), uuid_create() generated bad leading digit for patient SSN missing.\n");
		}
	}
}

void print_uuid_parms(FILE *stream, struct test_data *td) {
	fprintf(stream, "patient_name          %s\n", td->patient_name);
	if (td->patient_ssn) {
		fprintf(stream, "patient_ssn           %09d\n", td->patient_ssn);
	}
	fprintf(stream, "patient_gender        %d\n", td->patient_gender);
	fprintf(stream, "patient_birth_year    %d\n", td->patient_year);
	fprintf(stream, "patient_birth_month   %d\n", td->patient_month);
	fprintf(stream, "patient_birth_day     %d\n", td->patient_day);

	if (td->policy_name[0]) {
		fprintf(stream, "policy_name           %s\n", td->policy_name);
		if (td->policy_ssn) {
			fprintf(stream, "policy_ssn            %09d\n", td->policy_ssn);
		}
		fprintf(stream, "policy_gender         %d\n", td->policy_gender);
		fprintf(stream, "policy_birth_year     %d\n", td->policy_year);
		fprintf(stream, "policy_birth_month    %d\n", td->policy_month);
		fprintf(stream, "policy_birth_day      %d\n", td->policy_day);
	}
}


/******************************************************************************/
/* generate test records and dump them to the output file */
/******************************************************************************/

static void generate_data_set(const char *filename) {
	struct test_data td;
	FILE *stream;
	int counter;

	stream = fopen(filename, "w");
	if (stream == NULL) {
		printf("Unable to open output file '%s'.  Aborting.\n", filename);
		exit(1);
	}
	printf("Writing simulated patient records to output file '%s'.\n", filename);

	/* now actually generate the various records */
	counter = 1;
	init_test_data(&td);
	while(create_patient_record(&td)) {
		fprintf(stream, "data_record           %d\n", counter);
		print_uuid_parms(stream, &td);
		generate_and_print_uuid(stream, &td);
		fprintf(stream, "\n");

		init_test_data(&td);

		if (small_archive_generate && counter >= SMALL_ARCHIVE_RECORD_COUNT) {
			break;
		}

		counter++;
	}
}


/******************************************************************************/
/* main routine and usage */
/******************************************************************************/

int main(int argc, char *argv[]) {
	if (argc <= 1 || (argc == 2 && strcmp(argv[1], "-h") == 0)) {
		printf("This program creates archives of simulated patient records for\n");
		printf("testing and verification of the UUID generation algorithm and library.\n");
		printf("Options:\n");
		printf("\n");
		printf("%s -s             : create a small archive\n", argv[0]);
		printf("%s -l             : create a large, complete archive\n", argv[0]);
		printf("\n");
		printf("The small archive contains approximately twenty thousand records,\n");
		printf("in random date blocks grouped by SSN.\n");
		printf("\n");
		printf("For exhaustive or collision testing, the large version of the simulated\n");
		printf("data archive may be created using the '-l' option.  This archive is\n");
		printf("extremely large, and contains upwards of 15 million records.  It may\n");
		printf("take several minutes to generate and requires approximately 3 gb of\n");
		printf("disk space.\n");
		return 0;
	}

	if (argc == 2 && strcmp(argv[1], "-s") == 0) {
		small_archive_generate = 1;
		generate_data_set("simulated_patient_records_small.txt");
		return 0;
	}

	if (argc == 2 && strcmp(argv[1], "-l") == 0) {
		small_archive_generate = 0;
		generate_data_set("simulated_patient_records_large.txt");
		return 0;
	}

	printf("Sorry, unknown option.   See '%s -h' for options.\n", argv[0]);
	return 0;
}

