
/* ************************************************************************* */
/*
 *	util.c
 *
 *	(C) 2010 VOCAL Technologies, Ltd.
 *
 *	This software is intended for use as a reference/standard implementation,
 *	and has been placed into the public domain by the copyright holder:
 *
 *	VOCAL Technologies, Ltd.
 *	90A John Muir Drive.
 *	Buffalo, NY  14228
 *
 *	Maintainer:  Dennis Towne
 */
/* ************************************************************************* */

#include <string.h>

#include "system.h"
#include "util.h"


/******************************************************************************/
/* hex/unhex printing utilities */
/******************************************************************************/

static char hexchar(int c) {
	if(c <= 9)
		return '0'+c;
	if(c < 16)
		return 'a'+c-10;
	return ~0;
}

unsigned int unhex_char(int c) {
	if(c >= '0' && c <= '9')
		return c-'0';
	if(c >= 'A' && c <= 'F')
		return c-'A'+10;
	if(c >= 'a' && c <= 'f')
		return c-'a'+10;
	return 0xff;
}

char *hex(uchar *f, unsigned int n, char *t) {
	unsigned int i;
	char *o;

	o = t;
	for(i = 0; i < n; i++) {
		*t++ = hexchar(f[i] >> 4);
		*t++ = hexchar(f[i] & 0xf);
	}
	*t = 0;
	return o;
}

const char *unhex(const char *s, uchar *p, unsigned int n) {
	uchar a, b;
	unsigned int len;

	len = strlen(s);
	if(len > n*2)
		return "input too long";
	if(len % 2 != 0)
		return "input has odd length, bad hex";

	for(; len > 0 && n > 0; len -= 2, n--) {
		a = unhex_char(*s++);
		b = unhex_char(*s++);
		if(a == 0xff || b == 0xff) {
			return "bad hex character";
		}
		*p++ = (a<<4)|b;
	}
	return NULL;
}

